/***************************
FFT.cpp
par Silicium628
logiciel Libre Open Source
****************************/

#include <avr/io.h>
#include <stdint.h> 
#include <stdlib.h>
#include "timeout.h"   // comprend : #define F_CPU 16000000UL  // 16 MHz
#include "LCD_5110-628.cpp"
#include <math.h>


const char *version = "4.2";  // ( signal entrant 10kHz à 90kHz )

//typedef float complexe[2];
typedef int16_t complexe_f[2];  // entier signé pour calculs (rapides) en virgule fixe 

//complexe tab_X[64];  // nb_ech valeurs traitées
//complexe tab_W[32];  // nb_ech/2
complexe_f tab_X[64];  // nb_ech valeurs traitées
complexe_f tab_W[32];  // nb_ech/2

float frequence1; // Hz
float frq_echantillonnage; // Hz

const uint8_t nb_etapes=6; 
uint8_t nb_ech = pow (2,nb_etapes); // nombre d'échantillons = 2 puissance(nb_etapes)

char hamming;
char bourrage;
char mode_affi;


void RAZ_tableau_echantillons()
{
   uint8_t n;
   for(n=0; n < nb_ech; n++)
   {
       tab_X[n][1] = 0; // partie reelle
       tab_X[n][2] = 0; // partie imaginaire
   }
}


void init_ports (void)  // ports perso
// 0 = entree, 1=sortie   ;   les 1 sur les pins en entrees activent les R de Pull Up (tirage à VCC)
{
  PORTA = 0b00000111;
  DDRA  = 0b01111000;  //PA7 en entree (ADC7 - entrée signal); PA0, PA1, PA2 jumpers à GND.
  
  PORTB = 0b00000000;
  DDRB |= 0b11111111;
  
  PORTC = 0b00000000;
  DDRC =   0b11111111;

  PORTD = 0b00000000; 
  DDRD =  0b11111111;
}

/*
The ADC module contains a prescaler, which generates an acceptable ADC clock fre-
quency from any CPU frequency above 100 kHz. The prescaling is set by the ADPS bits
in ADCSRA
*/

void InitADC (void)
{
//  SFIOR  = 0b10000000; // ADTS2,1,0 = 100 -> CAN declenché par Timer0 Overflow - p:218
  ADMUX  = 0b01000111;   //Bit 7:6 = ADC Reference Selection = 01 -> AVCC ; //Bits 0:4 - Analog Channel Selection Bits et gain éventuel - p214
// ici Select pin ADC7 using MUX  avec ref tension interne = AREF externe - pas de gain.  

/** CHOISIR UNE SEULE LIGNE CI-DESSOUS pour ADCSRA */

// parametrage gamme 8-90 kHz
//  ADCSRA = 0b10000010;  // bit7: Activate ADC; ADPS2..0->Prescaler=f/4 ; bit5 = ADATE (ADC Auto Trigger Enable)->non utilisé ici - p:216 et 218

// parametrage gamme 4-56 kHz
//  ADCSRA = 0b10000011;  // bit7: Activate ADC; ADPS2..0->Prescaler=f/8 ; bit5 = ADATE (ADC Auto Trigger Enable)->non utilisé ici - p:216 et 218

// parametrage gamme 2-31 kHz
//  ADCSRA = 0b10000100;  // bit7: Activate ADC; ADPS2..0->Prescaler=f/16 ; bit5 = ADATE (ADC Auto Trigger Enable)->non utilisé ici - p:216 et 218
 
// parametrage gamme 1-15 kHz
//  ADCSRA = 0b10000101;  // bit7: Activate ADC; ADPS2..0->Prescaler=f/32 ; bit5 = ADATE (ADC Auto Trigger Enable)->non utilisé ici - p:216 et 218
 
// parametrage gamme 300 Hz-8000 Hz
//  ADCSRA = 0b10000110;  // bit7: Activate ADC; ADPS2..0->Prescaler=f/32 ; bit5 = ADATE (ADC Auto Trigger Enable)->non utilisé ici - p:216 et 218

// parametrage gamme 200 Hz- 4300 Hz
  ADCSRA = 0b10000111;  // bit7: Activate ADC; ADPS2..0->Prescaler=f/32 ; bit5 = ADATE (ADC Auto Trigger Enable)->non utilisé ici - p:216 et 218
 
}





uint8_t bit_reversal(uint8_t num, uint8_t nb_bits)
{
    uint8_t r = 0, i, s;
    if ( num > 1<< nb_bits) { return 0; }

    for (i=0; i<nb_bits; i++)
    {
        s = (num & (1 << i));
        if(s) { r |= (1 << ((nb_bits - 1) - i)); }
    }
    return r;
}



/***********************************************************************
Remarques :  
- si l'on veut augmenter la vitesse d'acquisition et maîtriser précisément
sa durée d'exécution il y a lieu de remplacer la fonction "acquisition()" par  
une routine écrite en assembleur et faire appel à une interruption pour le timing.
- si l'on veut de plus augmenter la vitesse du traitement de la FFT, il faudra
remplacer la fonction "calcul_fourier()" par une routine en assembleur, et en fait
dans ce cas on envisagera de réécrire tout le programme en assembleur.
- si on désire une plus grande rapidité encore, on pourra se tourner vers
un microcontrôleur ARM cadencé à... 700 MHz (voire plus) et on pense alors bien sûr au Raspberry Pi !
- si l'on veut calculer encore plus vite on peut utiliser le processeur d'une carte graphique,  
(par exemple en 2014 une Nvidia GeForce GTX680 dont le processeur tourne à à 2GHz- 2Go - DDR5 ou une Nvidia Quadro 6000 )
ou au moins le FPU intégré dans les microprocesseurs de nos ordinateurs.
- et si... on exige encore plus de vitesse on se tournera vers un Field Programmable Gate Arrays (FPGA) 
ou processeur de signal numérique (DSP) programmé en assembleur.
voir :
- http://fr.wikipedia.org/wiki/Processeur_de_signal_num%C3%A9rique
- http://www.orfony.fr/bdtech/DSP.html
- ADSP-TS001 (Analog Device)
- Stratix FPGA (Altera)
 
Toutefois cette implémentation sur un simple ATmega à 4 euros n'a pas la prétention
d’afficher des performances époustouflantes, mais elle permet de COMPRENDRE et de maîtriser 
la transformée de Fourier et de la voir fonctionner.
***********************************************************************/  

void acquisition()
{  
  uint16_t acqH;
  uint16_t tab_acq[64];
  uint8_t n, r, n_max;
  float x;
  uint16_t valeur_lue;  

  if (bourrage) {n_max = nb_ech /4; } else {n_max = nb_ech; }
    //n_max  = 16;
    
    for(n=0; n < n_max-1; n++)
    {    
        ADCSRA |= _BV(ADSC);    //Start conversion - resolution 10bits
        while (ADCSRA & _BV(ADSC) ) {;}    // attend la fin de la converstion
        tab_acq[n] = ADCL;
        acqH = ADCH; // passage à 16 bits pour pouvoir décaller
        tab_acq[n] += acqH << 8;
    }    
    for(n=0; n < n_max-1; n++)
    {
      valeur_lue =tab_acq[n];
      x= (valeur_lue - 512.0) / 100.0;   
      if (hamming)  { x = x * (1- cos (2* M_PI * n / n_max ));}       
      r=bit_reversal(n,nb_etapes);
      tab_X[r][1] = x * 16; // partie reelle  - * 16 pour obtenir un nombre en virgule fixe
      tab_X[r][2] = 0; // partie imaginaire
    }
}



void remplir_tableau_echantillons()
{
// création du signal à analyser
    uint8_t n, r, n_max;
    float Te; // période d'échantillonage
    float x;
    Te = 1.0 / frq_echantillonnage;

    if (bourrage==true) {n_max = nb_ech /2; } else {n_max = nb_ech; }

    for(n=0; n < n_max-1; n++) // nb d'échantillons
    {
        x=  cos(2.0 * M_PI * frequence1 * n * Te );
        if (hamming)  { x = x * (1- cos (2* M_PI * n / n_max ));}
        r=bit_reversal(n,nb_etapes);
        tab_X[r][1] = x; // partie reelle
        tab_X[r][2] = 0; // partie imaginaire
     }
}



void calcul_tableau_W()
{
// calcul et memorisation dans un tableau des twiddle factors
    uint8_t n;
    //float x;
    float x, cosx, sinx;

    for(n=0; n<(nb_ech/2-1); n++)
    {
        x=2.0*M_PI * n / nb_ech;
				cosx =  16 * cos(x);
				sinx = -16 * sin(x);

        tab_W[n][1] =  round(cosx);  // partie reelle
        tab_W[n][2] =  round(sinx);  // partie imaginaire
    }
}



void calcul_fourier()
{
    complexe_f produit; // voir le type "complexe" déclaré plus haut

    uint8_t etape, e1, e2, li, w, ci;
    uint8_t li_max=nb_ech;

    e2=1;

    for (etape=1; etape<=nb_etapes; etape++)
    {
        e1=e2; //(e1 evite d'effectuer des divisions e2/2 plus bas)
        e2=e2+e2;

        for (li=0; li<li_max; li+=1)
        {
            ci=li & (e2-1);  // ET bit à bit
            if (ci>(e1-1))
            {
                w=li_max/e2*(li & (e1 -1));  // ET bit à bit  calcul du numéro du facteur de tripatouillages W

                produit[1]=((tab_W[w][1]*tab_X[li][1]) - (tab_W[w][2]*tab_X[li][2])) >> 4; // le décallage ( /16) because multiplication en virgule fixe
                produit[2]=((tab_W[w][1]*tab_X[li][2]) + (tab_W[w][2]*tab_X[li][1])) >> 4;

                tab_X[li][1]=tab_X[li-e1][1]-produit[1];
                tab_X[li][2]=tab_X[li-e1][2]-produit[2];

                tab_X[li-e1][1]=tab_X[li-e1][1]+produit[1];
                tab_X[li-e1][2]=tab_X[li-e1][2]+produit[2];
            }
        }
    }
}



void tracer_signal()
{
    int32_t a, S;
    uint8_t  r,  memo_x, memo_y, n, x, y;

    x=0;
    y=24;
    for(n=0; n<32; n++)
    {
        memo_x = x;
        memo_y = y;
        x = 2*n;
        r=bit_reversal(n, nb_etapes);
        if (hamming ==0) {a = 10;} else {a=5;}
        S = (a * tab_X[r][1]); // partie réelle
        S =  S >> 4;
        S = 20 - S;
        if ( (abs(S) )< 48 )  {y= lrint(S);} else {y=1;}
        
/** choisir une des trois lignes ci-dessous **/    
        //buffer_drawPixel(x, y);    
        //buffer_trace_baton_V(x, y);
        buffer_trace_segment( memo_x, memo_y, x, y);
    }
    affi_buffer(84);
}



void tracer_tableau_valeurs()
{
  uint8_t offset_y = 0;
  uint8_t n ;
  uint8_t memo_x, memo_y, x, y;
  uint32_t A;


  uint8_t z;

    if (bourrage) {z=15;} else {z=5;}

    x=0;
    y=offset_y;
    
    for(n=0; n<31; n++)  // batons =3.5 ms ; segments = 86 ms
    {
        memo_x = x;
        memo_y = y;
        
        //x = n;        
        x = 2*n; // largeur de la trace = 4 px (1 baton de 3px + 1 séparation de 1px)
    
        A = ((tab_X[n][1] * tab_X[n][1]) + (tab_X[n][2] * tab_X[n][2]));
        A >>=8; // div/256; // 256 = 16 * 16 (compense la multiplication en virgule fixe puis retour à une valeur ordinaire)
        A *= z;
        A = sqrt(A);
			 
        if ( (abs(A) )< 48 )  {y= lrint(A);} else {y=47;}

/** choisir une des trois lignes ci-dessous; ATTENTION : les rapidités ne sont pas les mêmes !! **/    
      //  buffer_drawPixel(x, 47-y);    // la plus rapide, mais peu visible...
        buffer_trace_baton_V(x, 1 + y);
       // buffer_trace_segment( memo_x, memo_y, x, y);
    }

		affi_buffer(84); // 12.6 ms
   
}





int main(void)
{
  
  init_ports();
  InitADC();
//  InitINTs();
  
  _delay_ms(100);
  lcd_init();
    
  // frequence1 = 100.0; // Hz
  // frq_echantillonnage=3200.0; // Hz


  lcd_goto_LC(0,1);
  lcd_puts2("FTT ATmega32");

  _delay_ms(1000);
  lcd_clear();
      
  calcul_tableau_W();
   
  lcd_goto_LC(5,0);  
  if (! mode_affi) {lcd_puts2("TF Silicium628  ");} else {lcd_puts2(" Signal");}  // 7 ms  
  
    for(;;)
    {
      //frequence1 = 310.0; // Hz
            
      for(uint8_t n=1; n<23; n++)
      {
        mode_affi = (PINA & 0b00000001) == 0;
        hamming   = (PINA & 0b00000010) != 0;
        bourrage  = (PINA & 0b00000100) != 0;
  
        //frequence1 += 30;
        
      
				RAZ_tableau_echantillons();
				acquisition();  // 7ms
        
      //  remplir_tableau_echantillons();      
          
        if (mode_affi) 
        {
          buffer_RAZ();
          tracer_signal(); // entrée ATTENTION : il faut faire cet affichage AVANT de calculer la TF parce que les résultats écrasent le tableau d'entrée
        } 
        
        if (! mode_affi) 
        {  
          calcul_fourier();  // 7 ms avec calculs en virgule fixe (int 16 bits) au lieu de 100..120ms avec calculs en virgule flotante (float 32 bits)
          buffer_RAZ();    // 1.4 ms

				  tracer_tableau_valeurs(); // 16ms 
        }

      }
   }
}
